/*
===============================================================================

	FILE:	encoder.cpp
	
	CONTENTS:
	
		Face Fixer - SIGGRAPH 2000 demonstration software
	
	PROGRAMMERS:
	
		Martin Isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		28 April 2000 -- Martin - finalized for SIGGRAPH submission.
	
===============================================================================
*/
#include <iostream.h>

#include "encoder.h"

#include "mesh.h"
#include "vector.h"
#include "vertex.h"
#include "twinedge.h"
#include "codec.h"

FFencoder::FFencoder()
{
}

FFencoder::~FFencoder()
{
}

FFcodec* FFencoder::encode(FFmesh* mesh)
{
  int i;
  FFtwinEdge* run;

  FFcodec* codec = new FFcodec();

  int component_count = 0;

  for (i = 0; i<mesh->es->size(); i++)
  {
    run = (FFtwinEdge*)mesh->es->elementAt(i);
    run->component = 0;
  }

  mesh->gate = (FFtwinEdge*)mesh->es->elementAt(0);

  while (mesh->gate != 0)
  {
    int first_label = 1;
    FFtwinEdge *gate = mesh->gate;
    
    component_count++;

    gate->component = component_count;
    gate->boundary_next = gate->inv;
    gate->boundary_prev = gate->inv;
    
    gate->inv->component = component_count;
    gate->inv->boundary_next = gate;
    gate->inv->boundary_prev = gate;

    codec->pushGate(gate);

    while ( codec->hasMoreGates() )
      {
	gate = codec->popGate();
	
	FFtwinEdge* run = gate;
	do
	  {
	    run->gate = 0;
	    run = run->boundary_next;
	  } while (run != gate);
	
	while (1)
	  {
	    if (!(first_label) && (gate->boundary_prev == gate->boundary_next))
	      { // >>>>>>> label E <<<<<<<
		gate = 0;
		codec->pushLabelE();
		break;
	      }
	    else if (!(first_label) && (gate == gate->boundary_next->inv)) 
	      { // >>>>>>> label R <<<<<<<
		gate->boundary_next->boundary_next->boundary_prev = gate->boundary_prev;
		gate->boundary_prev->boundary_next = gate->boundary_next->boundary_next;
		gate = gate->boundary_prev;
		codec->pushLabelR();
	      }
	    else if (!(first_label) && (gate == gate->boundary_prev->inv))
	      { // >>>>>>> label L <<<<<<<
		gate->boundary_next->boundary_prev = gate->boundary_prev->boundary_prev;
		gate->boundary_prev->boundary_prev->boundary_next = gate->boundary_next;
		gate = gate->boundary_next;
		codec->pushLabelL();
	      }
	    else if (!(first_label) && (gate->inv->boundary_prev) && (gate->inv != gate->boundary_prev) && (gate->inv != gate->boundary_next) )
	      {
		if (gate->inv->gate == 0)
		  { // >>>>>>> label S <<<<<<<
		    FFtwinEdge* push_gate = gate->boundary_prev;
		    
		    gate->boundary_prev->boundary_next = gate->inv->boundary_next;
		    gate->inv->boundary_next->boundary_prev = gate->boundary_prev;
		    
		    gate->boundary_next->boundary_prev = gate->inv->boundary_prev;
		    gate->inv->boundary_prev->boundary_next = gate->boundary_next;
		    
		    codec->pushGate(push_gate);
		    run = push_gate;
		    do
		      {
			run->gate = push_gate;
			run = run->boundary_next;
		      } while (run != push_gate);
		    gate = gate->boundary_next;
		    codec->pushLabelS();
		  }
		else
		  { // >>>>>>> label M <<<<<<<
		    FFtwinEdge* stackgate = gate->inv->gate;
		    
		    int index = -1;
		    for (int i = 0; i < codec->gs->size(); i++)
		      {
			if (stackgate == (FFtwinEdge*)codec->gs->elementAt(i))
			  {
			    index = codec->gs->size() - i;
			    codec->gs->removeElementAt(i);
			    run = stackgate;
			    do
			      {
				run->gate = 0;
				run = run->boundary_next;
			      } while (run != stackgate);
			    break;
			  }
		      }
		    
		    if (index == -1)
		      {
			cerr << "stackgate not found in stack" << endl;
			cerr << "FATAL ERROR ... exiting" << endl;
			return 0;
		      }

		    run = gate->inv;
		    
		    int offset1 = 0;
		    while (run->boundary_prev != stackgate)
		      {
			offset1++;
			run = run->boundary_prev;
		      }
		    
		    int offset2 = 0;
		    while (run->boundary_prev->inv != gate)
		      {
			offset2++;
			run = run->boundary_prev;
		      }
		    
		    gate->boundary_prev->boundary_next = gate->inv->boundary_next;
		    gate->inv->boundary_next->boundary_prev = gate->boundary_prev;
		    
		    gate->boundary_next->boundary_prev = gate->inv->boundary_prev;
		    gate->inv->boundary_prev->boundary_next = gate->boundary_next;
		    
		    gate = gate->boundary_next;
		    codec->pushLabelM(index, offset1, offset2);
		  }
	      }
	    else
	      {
		gate->boundary_prev->boundary_next = gate->next;
		gate->next->boundary_prev = gate->boundary_prev;
		
		int degree = 2;
		run = gate;
		while (run->next->inv->next->inv != gate)
		  {
		    degree++;
		    run->next->component = component_count;
		    run->next->inv->component = component_count;
		    run = run->next->inv;
		    
		    run->inv->boundary_next = run->next;
		    run->next->boundary_prev = run->inv;
		  }
		
		run->next->component = component_count;
		run->next->inv->component = component_count;
		run->next->boundary_next = gate->boundary_next;
		gate->boundary_next->boundary_prev = run->next;
		
		if (gate->group == -1)
		  {
		    codec->pushLabelH(degree);
		  }		
		else
		  {
		    codec->pushLabelF(degree);
		  }
		gate = run->next;
	      }
	    first_label = 0;
	  }
      }
    
    // removes the very last E label which is not necessary
    codec->popLabel();

    // search for more mesh components to encode
    mesh->gate = 0;
    for (i = 0; i<mesh->es->size(); i++)
      {
	run = (FFtwinEdge*)mesh->es->elementAt(i);
	if ((run->component == 0) && (run->inv->component == 0))
	  {
	    mesh->gate = run;
	    break;
	  }
      }
  }
  
  mesh->NumComponents = component_count;
  mesh->NumHoles = codec->h;
  mesh->NumHandles = codec->m;
  cout << "# of components: " << component_count << endl;
  cout << "# of labels    : " << codec->ls->size() << endl;
  cout << " ...   f3 : " << codec->f[3] << endl;
  cout << " ...   f4 : " << codec->f[4] << endl;
  cout << " ...   f5 : " << codec->f[5] << endl;
  cout << " ...   f6 : " << codec->f[6] << endl;
  cout << " ...  f>6 : " << codec->large_f << endl;
  cout << " ...    r : " << codec->r << endl;
  cout << " ...    l : " << codec->l << endl;
  cout << " ...    s : " << codec->s << endl;
  cout << " ...    e : " << codec->e << endl;
  cout << " ...    h : " << codec->h << endl;
  cout << " ...    m : " << codec->m << endl;
  return codec;
}

FFmesh* FFencoder::decode(FFcodec* codec)
{
  FFmesh* mesh = new FFmesh();

  /*

    read the paper and fill in the missing lines ... (-:

   */

  return mesh;
}

void FFencoder::encodeStepwise(FFmesh* mesh, FFcodec* codec, int steps)
{
  int i;
  FFtwinEdge* run;

  if (steps == 0) return;

  while ( steps && (mesh->gate != 0) )
  {
    int first_label;
    FFtwinEdge *gate = mesh->gate;

    if ((gate->boundary_prev == 0) && (gate->boundary_next == 0))
      {
	first_label = 1;
	
	gate->boundary_next = gate->inv;
	gate->boundary_prev = gate->inv;
	
	gate->inv->boundary_next = gate;
	gate->inv->boundary_prev = gate;
      }
    else
      {
	first_label = 0;
      }

    codec->pushGate(gate);

    while ( steps && codec->hasMoreGates() )
      {
	gate = codec->popGate();
	
	FFtwinEdge* run = gate;
	do
	  {
	    run->gate = 0;
	    run = run->boundary_next;
	  } while (run != gate);
	
	while ( steps )
	  {
	    if (!(first_label) && (gate->boundary_prev == gate->boundary_next))
	      { // >>>>>>> label E <<<<<<<
		gate = 0;
		codec->pushLabelE();
		break;
	      }
	    else if (!(first_label) && (gate == gate->boundary_next->inv)) 
	      { // >>>>>>> label R <<<<<<<
		gate->boundary_next->boundary_next->boundary_prev = gate->boundary_prev;
		gate->boundary_prev->boundary_next = gate->boundary_next->boundary_next;
		gate = gate->boundary_prev;
		codec->pushLabelR();
	      }
	    else if (!(first_label) && (gate == gate->boundary_prev->inv))
	      { // >>>>>>> label L <<<<<<<
		gate->boundary_next->boundary_prev = gate->boundary_prev->boundary_prev;
		gate->boundary_prev->boundary_prev->boundary_next = gate->boundary_next;
		gate = gate->boundary_next;
		codec->pushLabelL();
	      }
	    else if (!(first_label) && (gate->inv->boundary_prev) && (gate->inv != gate->boundary_prev) && (gate->inv != gate->boundary_next) )
	      {
		if (gate->inv->gate == 0)
		  { // >>>>>>> label S <<<<<<<
		    FFtwinEdge* push_gate = gate->boundary_prev;
		    
		    gate->boundary_prev->boundary_next = gate->inv->boundary_next;
		    gate->inv->boundary_next->boundary_prev = gate->boundary_prev;
		    
		    gate->boundary_next->boundary_prev = gate->inv->boundary_prev;
		    gate->inv->boundary_prev->boundary_next = gate->boundary_next;
		    
		    codec->pushGate(push_gate);
		    run = push_gate;
		    do
		      {
			run->gate = push_gate;
			run = run->boundary_next;
		      } while (run != push_gate);
		    gate = gate->boundary_next;
		    codec->pushLabelS();
		  }
		else
		  { // >>>>>>> label M <<<<<<<
		    FFtwinEdge* stackgate = gate->inv->gate;
		    
		    int index = -1;
		    for (int i = 0; i < codec->gs->size(); i++)
		      {
			if (stackgate == (FFtwinEdge*)codec->gs->elementAt(i))
			  {
			    index = codec->gs->size() - i;
			    codec->gs->removeElementAt(i);
			    run = stackgate;
			    do
			      {
				run->gate = 0;
				run = run->boundary_next;
			      } while (run != stackgate);
			    break;
			  }
		      }
		    
		    if (index == -1)
		      {
			cerr << "stackgate not found in stack" << endl;
			cerr << "FATAL ERROR ... exiting" << endl;
			return;
		      }

		    run = gate->inv;
		    
		    int offset1 = 0;
		    while (run->boundary_prev != stackgate)
		      {
			offset1++;
			run = run->boundary_prev;
		      }
		    
		    int offset2 = 0;
		    while (run->boundary_prev->inv != gate)
		      {
			offset2++;
			run = run->boundary_prev;
		      }
		    
		    gate->boundary_prev->boundary_next = gate->inv->boundary_next;
		    gate->inv->boundary_next->boundary_prev = gate->boundary_prev;
		    
		    gate->boundary_next->boundary_prev = gate->inv->boundary_prev;
		    gate->inv->boundary_prev->boundary_next = gate->boundary_next;
		    
		    gate = gate->boundary_next;
		    codec->pushLabelM(index, offset1, offset2);
		  }
	      }
	    else
	      {
		gate->boundary_prev->boundary_next = gate->next;
		gate->next->boundary_prev = gate->boundary_prev;
		
		int degree = 2;
		run = gate;
		run->marker = 1;
		while (run->next->inv->next->inv != gate)
		  {
		    degree++;
		    run = run->next->inv;
		    run->marker = 1;

		    run->inv->boundary_next = run->next;
		    run->next->boundary_prev = run->inv;
		  }
		run->next->inv->marker = 1;
		run->next->boundary_next = gate->boundary_next;
		gate->boundary_next->boundary_prev = run->next;
		
		if (gate->group == -1)
		  {
		    codec->pushLabelH(degree);
		  }		
		else
		  {
		    codec->pushLabelF(degree);
		  }
		gate = run->next;
	      }
	    first_label = 0;
	    steps--;
	  }
      }
    
    if ( (gate == 0) && (codec->hasMoreGates() == 0))
      {
	// removes the very last E label which is not necessary
	codec->popLabel();

        // search for more mesh components to encode
	mesh->gate = 0;
	for (i = 0; i<mesh->es->size(); i++)
	  {
	    run = (FFtwinEdge*)mesh->es->elementAt(i);
	    if ((run->boundary_prev == 0) && (run->inv->boundary_next == 0))
	      {
		mesh->gate = run;
		break;
	      }
	  }
      }
    else
    {
      mesh->gate = gate;
    }
  }
}
