/*
===============================================================================

	FILE:	vec3fv.cpp
	
	CONTENTS:
	
		Face Fixer - SIGGRAPH 2000 demonstration software
	
	PROGRAMMERS:
	
		Martin Isenburg@cs.unc.edu  (with ideas stolen from Kenny Hoff)
	
	COPYRIGHT:
	
		Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		28 April 2000 -- Martin - finalized for SIGGRAPH submission.
	
===============================================================================
*/
#include "vec3fv.h"
#include <math.h>

void VecSet3fv(float v[3], float x, float y, float z)
{
  v[0] = x;
  v[1] = y;
  v[2] = z;
}

void VecCopy3fv(float v[3], float a[3])
{
  v[0] = a[0];
  v[1] = a[1];
  v[2] = a[2];
}

void VecSelfScalarMult3fv(float v[3], float s)
{
  v[0] *= s;
  v[1] *= s;
  v[2] *= s;
}

void VecSelfScalarDiv3fv(float v[3], float s)
{
  v[0] /= s;
  v[1] /= s;
  v[2] /= s;
}

void VecSelfNegate3fv(float v[3])
{
  v[0] = -v[0];
  v[1] = -v[1];
  v[2] = -v[2];
}

void VecSelfAdd3fv(float v[3], float a[3])
{
  v[0] += a[0];
  v[1] += a[1];
  v[2] += a[2];
}

void VecSelfSubtract3fv(float v[3], float a[3])
{
  v[0] -= a[0];
  v[1] -= a[1];
  v[2] -= a[2];
}

void VecAdd3fv(float c[3], float a[3], float b[3])
{
  c[0] = a[0] + b[0];
  c[1] = a[1] + b[1];
  c[2] = a[2] + b[2];
}

void VecSubtract3fv(float c[3], float a[3], float b[3])
{
  c[0] = a[0] - b[0];
  c[1] = a[1] - b[1];
  c[2] = a[2] - b[2];
}

float VecLength3fv(float v[3])
{
  return( (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]) );
}

void VecNormalize3fv(float v[3])
{
  float l = VecLength3fv(v);
  v[0] /= l;
  v[1] /= l;
  v[2] /= l;
}

float VecDotProd3fv(float a[3], float b[3])
{
  return( a[0]*b[0] + a[1]*b[1] + a[2]*b[2] );
}

void VecCrossProd3fv(float c[3], float a[3], float b[3]) // c = a X b
{ 
  VecSet3fv(c, a[1]*b[2]-a[2]*b[1], a[2]*b[0]-a[0]*b[2], a[0]*b[1]-a[1]*b[0]);
}

void VecUpdateMinMax3fv(float min[3], float max[3], float v[3])
{
  if (v[0]<min[0]) min[0]=v[0]; else if (v[0]>max[0]) max[0]=v[0];
  if (v[1]<min[1]) min[1]=v[1]; else if (v[1]>max[1]) max[1]=v[1];
  if (v[2]<min[2]) min[2]=v[2]; else if (v[2]>max[2]) max[2]=v[2];
}

int VecEqual3fv(float a[3], float b[3])
{
  return( a[0]==b[0] && a[1]==b[1] && a[2]==b[2] );
}

void VecCcwNormal3fv(float n[3], float a[3], float b[3], float c[3])
{
  float ab[3], ac[3];
  VecSubtract3fv(ab,b,a);
  VecSubtract3fv(ac,c,a);
  VecCrossProd3fv(n,ab,ac);
  VecNormalize3fv(n);
}
